
classdef Admin % in Admin.m
    methods (Static = true)
        
        function [S_l,S_l_c]= Compute_S_l_matrix(asset_group_vec)
            % compute selection matrix N x g
            % put a 1 when asset i (i = 1,..N) belongs to group j (j =
            % 1,...g)
            
            nr_groups = max(asset_group_vec);
            n   = length(asset_group_vec);
            
            S_l = zeros(n,nr_groups);
            
            for i = 1:n
                S_l(i,asset_group_vec(i)) = 1;
            end
            
            S_l_c = ones(n,nr_groups) - S_l;
            
        end
        
        function S_d_sp= Compute_S_d_matrix(asset_group_vec)
            % create a N^2 by g selection matrix (with g the number of
            % groups)
            % such that entry (i + (i-1)*N, j) = 1 if asset i belongs to
            % group j and 0 else  (j = 1,..g)  (i = 1,..n)
            
            nr_groups   = max(asset_group_vec);
            n           = length(asset_group_vec);
            S_d         = zeros(n^2,nr_groups);
            
            for i = 1:n
                S_d(i + (i-1)*n,asset_group_vec(i)) = 1;
            end
            S_d_sp = sparse(S_d);
            
        end
        
        function R_fit = Compute_R_fit_given_L_and_D(L_mat,D_mat)
            
            [T,k] = size(L_mat);
            R_fit = zeros(k,k,T);
            
            for i = 1:T
                R_fit(:,:,i) = L_mat(i,:)'*L_mat(i,:) + diag(D_mat(i,:));
            end
            
        end
        
        function R_t = Compute_R_t(sigma_2_vec_t,L_til_prime_t)
            % Compute R_t following eq (4) of Opschoor et al. 2020
            
            R_t = L_til_prime_t*L_til_prime_t' + diag(sigma_2_vec_t);
            
        end
        
        
        function [R_inv_t,det_R_t,temp_0,det_D_t] = Compute_R_t_inv_and_det_R_t(sigma_2_vec_t,L_til_prime_t)
            % Compute inverse of R_t and det R_t using eq (5) of Opschoor et al (2020)
            
            sigma_2_inv           = 1./sigma_2_vec_t;
            L_til_t_times_D_inv_t = L_til_prime_t.*sigma_2_inv;
            temp_0                = 1 + L_til_t_times_D_inv_t'*L_til_prime_t;
            D_inv_t               = diag(sigma_2_inv);
            
            R_inv_t = D_inv_t - (1/temp_0)*(L_til_t_times_D_inv_t* L_til_t_times_D_inv_t');
            det_D_t =  prod(sigma_2_vec_t);
            det_R_t = temp_0*det_D_t;
            
        end
        
        function R_inv_t = Compute_R_t_inv(sigma_2_vec_t,L_til_prime_t)
            % Compute inverse of R_t using eq (5) of Opschoor et al (2020)
            
            sigma_2_inv           = 1./sigma_2_vec_t;
            L_til_t_times_D_inv_t = L_til_prime_t.*sigma_2_inv;
            temp_0                = 1 + L_til_t_times_D_inv_t'*L_til_prime_t;
            D_inv_t               = diag(sigma_2_inv);
            
            R_inv_t = D_inv_t - (1/temp_0)*(L_til_t_times_D_inv_t* L_til_t_times_D_inv_t');
            
        end
        
        
        function [R_inv_t,det_R_t,Flag] = Compute_R_t_inv_and_det_R_t_L_matrix(sigma_2_vec_t,L_til_prime_t)
            % Compute inverse of R_t and det R_t using eq (5) of Opschoor et al (2020)
            
            Flag    = 0;
            m       = size(L_til_prime_t,2);
            
            sigma_2_inv           = 1./sigma_2_vec_t;
            L_til_t_times_D_inv_t = L_til_prime_t.*(sigma_2_inv*ones(1,m));
            temp_0                = eye(m) +  L_til_t_times_D_inv_t'*L_til_prime_t;
            
            [~,ind_pd] = chol(temp_0);
            if ind_pd>0
                Flag = 1;
                R_inv_t = nan;
                det_R_t = nan;
            else
                D_inv_t = diag(1./sigma_2_vec_t);
                
                R_inv_t = D_inv_t - (L_til_t_times_D_inv_t * inv(temp_0))*L_til_t_times_D_inv_t';
                det_D_t =  prod(sigma_2_vec_t);
                det_R_t = det(temp_0)*det_D_t;
            end
            
        end
        
        function [R_inv_t,det_R_t,Flag] = Compute_R_t_inv_L_matrix(sigma_2_vec_t,L_til_prime_t)
            % Compute inverse of R_t using eq (5) of Opschoor et al (2020)
            
            Flag = 0;
            m    = size(L_til_prime_t,2);
            sigma_2_inv           = 1./sigma_2_vec_t;
            L_til_t_times_D_inv_t = L_til_prime_t.*(sigma_2_inv*ones(1,m));
            temp_0                = eye(m) +  L_til_t_times_D_inv_t'*L_til_prime_t;
            
            [~,ind_pd] = chol(temp_0);
            if ind_pd>0
                Flag    = 1;
                R_inv_t = nan;
                det_R_t = nan;
            else
                D_inv_t = diag(1./sigma_2_vec_t);
                
                R_inv_t = D_inv_t - (L_til_t_times_D_inv_t * inv(temp_0))*L_til_t_times_D_inv_t';
            end
            
        end
        
        
        function [Rho_block_mat,Rt_Block] = Compute_BLOCK_correlation_matrix(x_mat, n_vec)
            % build a G times G block correlatoin matrix where
            % entry (i,j) (i not equal to j) means the average beween
            % correlation between industry i and j
            % and entry (i,i) indicates the average WITHIN correlation of
            % industry i.
            
            [T,N] = size(x_mat);
            G = length(n_vec);
            Rt_mat = corr(x_mat);
            
            teller_i = 1;
            Rho_block_mat = zeros(G,G);
            Rt_0 = zeros(N,N);
            Rt_1 = zeros(N,N);
            % loop over all G industries and compute the average WITHIN and BETWEEN
            % correlatinos.
            for i = 1:G
                n_i = n_vec(i);
                Rt_mat_ii = Rt_mat(teller_i:teller_i+n_i-1,teller_i:teller_i+n_i-1);
                if n_i>1
                    rho_ii = (ones(n_i,1)'*Rt_mat_ii*ones(n_i,1) - n_i)/(n_i*(n_i-1));
                else
                    rho_ii = 0;
                end
                Rho_block_mat(i,i) = rho_ii;
                
                Rt_0(teller_i:teller_i+n_i-1,teller_i:teller_i+n_i-1) = (1- rho_ii)*eye(n_i);
                Rt_1(teller_i:teller_i+n_i-1,teller_i:teller_i+n_i-1) = rho_ii;
                
                teller_j = sum(n_vec(1:i))+1;
                for b = i+1:G
                    Rt_mat_ij = Rt_mat(teller_i:teller_i+n_i-1,teller_j:teller_j+n_vec(b)-1);
                    rho_ij = mean(Rt_mat_ij(:));
                    Rho_block_mat(i,b) = rho_ij;
                    Rho_block_mat(b,i) = rho_ij;
                    
                    Rt_1(teller_i:teller_i+n_i-1,teller_j:teller_j+n_vec(b)-1) = rho_ij;
                    Rt_1(teller_j:teller_j+n_vec(b)-1,teller_i:teller_i+n_i-1) = rho_ij;
                    
                    teller_j = teller_j+n_vec(b);
                end
                
                teller_i = teller_i+n_vec(i);
            end
            
            Rt_Block  = Rt_0 + Rt_1;
        end
        
        function [se_H,se_S,se_G] = Compute_se_errors_ML(f_handle,theta_opt,T)
            
            % step 1: compute hessian
            H = hessian_2sided(f_handle,theta_opt);
            inv_H    = inv(H);
            se_H = sqrt(diag(inv_H));
            
            % step 2 Outer product of gradient
            OPG_mat = Compute_OPG_MV_MODELS(f_handle,theta_opt,T);
            S_G = inv(OPG_mat)/T;
            se_G = sqrt(diag(S_G));
            
            % step 3; build sandwich covariance matrix
            S_COV = T * inv_H * OPG_mat * inv_H;
            se_S = sqrt(diag(S_COV));
        end
        
        function OPG_mat = Compute_OPG_MV_MODELS(f_handle, theta_vec0,T)
            % compute the expected value of the outer product of the gradient
            % associated wit the log-likelihood function (given by
            % f_handle)
            
            dEps=eps(1);
            dH=dEps^(1/3)*max( abs(theta_vec0), 1e-2);
            k = length(theta_vec0);
            g_mat = zeros(T,k);
            
            for i = 1:k
                
                theta_vec = theta_vec0;
                
                theta_vec(i) = theta_vec0(i) + dH(i);
                [~,LogLike] = f_handle(theta_vec);
                L_plus = -LogLike;
                
                theta_vec(i) = theta_vec0(i) - dH(i);
                [~,LogLike] = f_handle(theta_vec);
                L_min = -LogLike;
                
                g_mat(:,i) = (L_plus - L_min)/ (2 * dH(i));
                
            end
            
            OPG_mat = zeros(k,k);
            for j = 1:T
                OPG_mat = OPG_mat + g_mat(j,:)'*g_mat(j,:);
            end
            
            OPG_mat = OPG_mat/T;
            
        end
        
        %%
        function H = hessian_2sided(f,x,varargin)
            % PURPOSE:
            %      Computes 2-sided finite difference Hessian
            %
            % USAGE:
            %      H = hessian_2sided(func,x,varargin)
            %
            % INPUTS:
            %      func         - function name, fval = func(x,varargin)
            %      x            - vector of parameters (n x 1)
            %      varargin     - optional arguments passed to the function
            %
            % OUTPUTS:
            %      H            - finite differnce, 2-sided hessian
            %
            % COMMENTS:
            %      Code originally from COMPECON toolbox [www4.ncsu.edu/~pfackler]
            %      documentation modified to fit the format of the Ecoometrics Toolbox
            %      by James P. LeSage, Dept of Economics
            %      University of Toledo
            %      2801 W. Bancroft St,
            %      Toledo, OH 43606
            %      jlesage@spatial-econometrics.com
            %
            % Further modified (to do 2-sided numerical derivs, rather than 1) by:
            % Author: Kevin Sheppard
            % kevin.sheppard@economics.ox.ac.uk
            % Revision: 3    Date: 4/1/2004
            
            try
                feval(f,x,varargin{:});
            catch
                error('There was an error evaluating the function.  Please check the arguements.');
            end
            
            
            n = size(x,1);
            fx = feval(f,x,varargin{:});
            
            % Compute the stepsize (h)
            
            h = eps.^(1/3)*max(abs(x),1e-2);
            xh = x+h;
            h = xh-x;
            ee = sparse(1:n,1:n,h,n,n);
            
            % Compute forward and backward steps
            gp = zeros(n,1);
            for i=1:n
                gp(i) = feval(f,x+ee(:,i),varargin{:});
            end
            gm = zeros(n,1);
            for i=1:n
                gm(i) = feval(f,x-ee(:,i),varargin{:});
            end
            
            
            H=h*h';
            Hm=H;
            Hp=H;
            % Compute "double" forward and backward steps
            for i=1:n
                for j=i:n
                    Hp(i,j) = feval(f,x+ee(:,i)+ee(:,j),varargin{:});
                    Hp(j,i)=Hp(i,j);
                    Hm(i,j) = feval(f,x-ee(:,i)-ee(:,j),varargin{:});
                    Hm(j,i)=Hm(i,j);
                end
            end
            
            %Compute the hessian
            for i=1:n
                for j=i:n
                    H(i,j) = (Hp(i,j)-gp(i)-gp(j)+fx+fx-gm(i)-gm(j)+Hm(i,j))/H(i,j)/2;
                    H(j,i) = H(i,j);
                end
            end
        end
        
        
        
        function [uLT] = UnitLT(k)
            % fill Lower triangular matrix with ones
            uLT= zeros(k,k);
            
            for j=1:k
                for i=j+1:k
                    uLT(i,j) =1;
                end
            end
            
        end
        
        
        
        %%
        function [ A ] = Vech2Sym(k, vechA )
            % this function puts vech of A into A
            
            A= zeros(k,k);
            l=1;
            for j=1:k
                for i=j:k
                    A(i,j)= vechA(l);
                    A(j,i)= A(i,j);
                    l=l+1;
                end
            end
            
        end
        
        function [ A ] = Vec2Sym(k, vecA )
            % this function puts vec of A into A
            
            A= zeros(k,k);
            l=1;
            for j=1:k
                for i=1:k
                    A(i,j)= vecA(l);
                    l=l+1;
                end
            end
            
        end
        
        
        function [ A ] = VecCorr2Sym(k, vecA )
            % this function puts vec of LT mat of A into A
            
            A= zeros(k,k);
            l=1;
            for j=1:k
                for i=j+1:k
                    A(i,j)= vecA(l);
                    l=l+1;
                end
            end
            
        end
        
        
        
        %%
        function [ vechA ] = Sym2Vech(k, A )
            % this function puts A into vech of A
            
            vechA= zeros(k*(k+1)/2, 1);
            l=1;
            for j=1:k
                for i=j:k
                    vechA(l) = A(i,j);
                    l=l+1;
                end
            end
            
        end
        
        %%
        function [ vechltA ] = LowerTr2Vech(k, ltA )
            % this function puts lower triangular A into  vech lt A
            
            vechltA= zeros(k*(k+1)/2, 1);
            l=1;
            for j=1:k
                for i=j:k
                    vechltA(l)= ltA(i,j);
                    l=l+1;
                end
            end
            
        end
        
        function [UtA] =Sym2UT(k,A)
            % this function puts a symmetrix matrix A in to a
            % Upper triangular matrix
            UtA = zeros(k*(k-1)/2,1);
            l = 1;
            for j = 1:k
                UtA(l:l+k-j-1,1) = A(j,j+1:end);
                l =l+k-j;
            end
            
        end
        
        function [A] =Ut2Sym(k,UtA)
            % Upper triangular of A into A with ones on the diagonal
            A = eye(k);
            l = 1;
            for j = 1:k
                A(j,j+1:end) = UtA(l:l+k-j-1,1);
                A(j+1:end,j) =  A(j,j+1:end);
                l =l+k-j;
            end
            
        end
        
        
        
        
        %%
        function [ ltA ] = Vech2LowerTr(k, vechltA )
            % this function puts vector of vech of lt A into lt A
            
            ltA= zeros(k,k);
            l=1;
            for j=1:k
                for i=j:k
                    ltA(i,j)= vechltA(l);
                    l=l+1;
                end
            end
            
        end
        
        %%
        function [K_sp] = commutation(m,  n)
            
            I= eye(m*n, m*n);
            K= zeros(m*n, m*n);
            
            for j=1:n
                for i=1:m
                    K((i-1)*n + j,:)= I((j-1)*m + i, :);
                end
            end
            
            K_sp = sparse(K);
            
        end
        
        
        %%
        function [ D ] = duplication(k)
            % creates duplication matrix
            
            I= eye(k*(k+1)/2);
            D= zeros( k^2, k*(k+1)/2 );
            
            for j=1:k*(k+1)/2
                for i=1:k
                    if (i>=j)
                        D((j-1)*k + i ,:)= I((j-1)*(k - j/2) +i ,:);
                        D((i-1)*k + j ,:)= I((j-1)*(k - j/2) +i ,:);
                    end
                end
            end
            
        end
        
        %%
        function [ L ] = elimination(k)
            % creates elimination matrix
            
            D= Admin.duplication(k);
            L= inv(D' * D) * D';
            
        end
        
        
        %%
        function [DL] = duplication_gen(k)
            
            D = Admin.duplication(k);
            LT= ones(k, k);
            
            for i=1:k
                for j=i+1:k
                    LT(i,j)= 0;
                end
            end
            
            LT= reshape(LT, k^2, 1) * ones(1, k*(k+1)/2 );
            DL= LT .* D;
            
        end
        
        %%
        function [LL] = elimination_gen(k)
            
            DL= Admin.duplication_gen(k);
            LL= DL';
            
        end
    end % end methods
end % end class