% MAIN FILE
% date: April 2020

% This file imports simulated data from the csv file 'Testdata.csv', which
% contains PITs (1000 by 100 matrix), sorted across columns such that
% columns (i.e. assets) (g-1)10+1 : g*10 belong to group g (g = 1,...,10).

% Then the file estimates parameters of various single and/or multi factor copula models
% proposed in "Closed-Form Multi-Factor Copula Models with Observation-Driven Dynamic
% Factor Loadings" by Opschoor et al. (2020).
% The parameters are stored in the cell array 'para_cell'
% If required, the file alse computes the standard errors using the
% sandwich estimator of the covariance matrix.

% Estimated parameters of the 1F t, MF t and MF-LT t models can be found in the README file.

% See the README file for more information
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% import data and gather T and N
u_mat      = readmatrix('Testdata.csv');
u_mat(:,1) = [];
[T,N]      = size(u_mat);

% put to if you would like to compute standard errors based on the sandwich
% covariance matrix
ind_se     = 1;

% information about group size and group allocation
n_vec           = 10*ones(10,1);            % number of assets per group
asset_group_vec = kron([1:10]',ones(10,1));
G               = max(asset_group_vec);     %number of groups

% Each number in the vector model_est_ind_vec represents a unique factor copula model:
% 1 (2) 1F eq N (t)
% 3 (4) 1F gr N (t)
% 5 (6) 2F N (t)
% 7 (8) MF N (t)
% 9 (10) MF-LT N (t)

model_est_ind_vec = [2 4 10];
nr_models         = length(model_est_ind_vec);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%% Matlab routine to set up constrained optimization %%%%%%%%

options = optimset('fmincon');
options = optimset(options, 'Display', 'off');
options = optimset(options, 'Diagnostics', 'off');
options = optimset(options, 'Algorithm', 'sqp');
options.MaxIter     = 30000;
options.MaxFunEvals = 30000;

% Create starting values and lower/upper bounds for the contrained optimization
para_start_1f_eq  = [0.001 0.01 0.92 30]';
para_start_1f_gr  = [0.001*ones(1,G) 0.02 0.9 30]';
para_start_2f = [0.001*ones(1,G) 0.02 0.01*ones(1,2) 0.9 30]';
para_start_Mf = [0.05 0.02*ones(1,G) 0.01*ones(1,2) 0.9 30]';

% lower and upper bounds
Lb_1f_eq = [-1 eps*ones(1,2) 2.1]';
Rb_1f_eq = [ 0.1 0.10 0.99999 5000]';
Lb_1f_gr = [-1*ones(1,G) eps*ones(1,2) 2.1]';
Rb_1f_gr = [0.1*ones(1,G) 0.10 0.99999 5000]';
Lb_2f = [-1*ones(1,G+1) eps*ones(1,3) 2.1]';
Rb_2f = [0.1*ones(1,G+1) 0.1 0.1 0.99999 5000]';
Lb_Mf = [ eps*ones(1,4+G) 2.1]';
Rb_Mf = [0.5*ones(1,G+1) 0.2 0.2 0.99999 5000]';

nr_l  = G*(G+1)/2;
ind_pos =  [1; cumsum(G:-1:2)' + 1];

% Lower and upper bounds for the first step moment estimator of MF_LT
Lb_Mf_LT_1 = -3*ones(nr_l,1);
Lb_Mf_LT_1(ind_pos) = eps;
Rb_Mf_LT_1 =  3*ones(nr_l,1);
para_start_Mf_LT_1 = 0.2 * ones(nr_l,1);

% Lower and upper bounds for the second step ML estimator of MF_LT
Lb_Mf_LT_2 = [eps*ones(1,2) 2.1]';
Rb_Mf_LT_2 = [5*ones(1,1) 0.99999 5000]';
para_start_Mf_LT_2 = [0.01 0.96 20]';

% create a empty cell array for parameters, standard errors (optional) and strings (will be used later on to print the
% results)
para_cell = cell(10,1);
if ind_se==1
    se_para_cell = cell(10,1);
end
para_cell_str = cell(10,1);
para_1f_eq_str = {'omega', 'A','B', 'nu Cop'};
para_1f_gr_str = {'omega_1','omega_2','omega_3','omega_4','omega_5','omega_6','omega_7','omega_8','omega_9','omega_10', 'A','B', 'nu Cop'};
para_2f_str = {'omega_eq','omega_1','omega_2','omega_3','omega_4','omega_5','omega_6','omega_7','omega_8','omega_9','omega_10', 'A_eq','A_gr','B', 'nu Cop'};
para_Mf_str = para_2f_str;
para_Mf_LT_str = {'A','B', 'nu Cop'};

para_cell_str{1} = para_1f_eq_str(1:end-1);
para_cell_str{2} = para_1f_eq_str;
para_cell_str{3} = para_1f_gr_str(1:end-1);
para_cell_str{4} = para_1f_gr_str;
para_cell_str{5} = para_2f_str(1:end-1);
para_cell_str{6} = para_2f_str;
para_cell_str{7} = para_Mf_str(1:end-1);
para_cell_str{8} = para_Mf_str;
para_cell_str{9} = para_Mf_LT_str(1:end-1);
para_cell_str{10} = para_Mf_LT_str;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Estimate the model parameters in three steps:
% 1) Define the function (handle)
% 2) Minimize the MINUS loglikelihood using fmincon
% 3) Store the estimated parameters, loglikelihood and Exit Flag in a cell-array
% 4) (OPTIONAL: compute the standard errors and store these in a
% cell-array)

% NB: Flag= 2 means that we have found a local optimum


for i = 1:nr_models
    
    model_i = model_est_ind_vec(i);
    
    switch model_i
        
        case 1
            % 1f-equi N model
            f_Cop_1f_eq_N = @(params)LogLik_Factor_Copula.LogLik_Copula_1f_eq(N,T,params,u_mat,0);
            [theta_1f_eq_N, fval_1f_eq_N, Flag_1f_eq_N] = fmincon(f_Cop_1f_eq_N,para_start_1f_eq(1:end-1),[],[],[],[],Lb_1f_eq(1:end-1),Rb_1f_eq(1:end-1),[],options);
            para_cell{model_i} = [theta_1f_eq_N' -fval_1f_eq_N Flag_1f_eq_N];
            
            % used to print the results
            f_handle  = f_Cop_1f_eq_N;
            theta_opt = theta_1f_eq_N;
            max_LogLik = -fval_1f_eq_N;
            
        case 2
            % 1f-equi t model
            f_Cop_1f_eq_t = @(params)LogLik_Factor_Copula.LogLik_Copula_1f_eq(N,T,params,u_mat,1);
            [theta_1f_eq_t, fval_1f_eq_t, Flag_1f_eq_t] = fmincon(f_Cop_1f_eq_t, para_start_1f_eq,[],[],[],[],Lb_1f_eq,Rb_1f_eq,[],options);
            para_cell{model_i} = [theta_1f_eq_t' -fval_1f_eq_t Flag_1f_eq_t];
            
            % used to print the results
            f_handle  = f_Cop_1f_eq_t;
            theta_opt = theta_1f_eq_t;
            max_LogLik = -fval_1f_eq_t;
            
        case 3
            % 1f gr N model
            f_Cop_1f_gr_N = @(params)LogLik_Factor_Copula.LogLik_Copula_1f_gr(N,T,params,u_mat,asset_group_vec,0,0);
            tic;[theta_1f_gr_N, fval_1f_gr_N, Flag_1f_gr_N] = fmincon(f_Cop_1f_gr_N,para_start_1f_gr(1:end-1) ,[],[],[],[],Lb_1f_gr(1:end-1),Rb_1f_gr(1:end-1),[],options);toc
            para_cell{model_i} = [theta_1f_gr_N' -fval_1f_gr_N Flag_1f_gr_N];
            
            % used to print the results
            f_handle  = f_Cop_1f_gr_N;
            theta_opt = theta_1f_gr_N;
            max_LogLik = -fval_1f_gr_N;
            
        case 4
            % 1f gr t model
            f_Cop_1f_gr_t = @(params)LogLik_Factor_Copula.LogLik_Copula_1f_gr(N,T,params,u_mat,asset_group_vec,1,0);
            [theta_1f_gr_t, fval_1f_gr_t, Flag_1f_gr_t] = fmincon(f_Cop_1f_gr_t, para_start_1f_gr,[],[],[],[],Lb_1f_gr,Rb_1f_gr,[],options);
            para_cell{model_i} = [theta_1f_gr_t' -fval_1f_gr_t Flag_1f_gr_t];
            
            % used to print the results
            f_handle  = f_Cop_1f_gr_t;
            theta_opt = theta_1f_gr_t;
            max_LogLik = -fval_1f_gr_t;
            
        case 5
            % 2f N model
            f_Cop_2f_N = @(params)LogLik_Factor_Copula.LogLik_Copula_2f(N,T,params,u_mat,asset_group_vec,0,0);
            [theta_Cop_2f_N, fval_2f_N, Flag_2f_N] = fmincon(f_Cop_2f_N,para_start_2f(1:end-1) ,[],[],[],[],Lb_2f(1:end-1),Rb_2f(1:end-1),[],options);
            para_cell{model_i} = [theta_Cop_2f_N' -fval_2f_N Flag_2f_N];
            
            % used to print the results
            f_handle  = f_Cop_2f_N;
            theta_opt = theta_2f_N;
            max_LogLik = -fval_2f_N;
            
        case 6
            % 2f t model
            f_Cop_2f_t = @(params)LogLik_Factor_Copula.LogLik_Copula_2f(N,T,params,u_mat,asset_group_vec,1,0);
            tic;[theta_Cop_2f_t, fval_2f_t, Flag_2f_t] = fmincon(f_Cop_2f_t,para_start_2f ,[],[],[],[],Lb_2f,Rb_2f,[],options);toc
            para_cell{model_i} = [theta_Cop_2f_t' -fval_2f_t Flag_2f_t];
            
            % used to print the results
            f_handle  = f_Cop_2f_t;
            theta_opt = theta_2f_t;
            max_LogLik = -fval_2f_t;
            
        case 7
            % MF N model
            f_Cop_Mf_N = @(params)LogLik_Factor_Copula.LogLik_Copula_MF(N,T,params,u_mat,asset_group_vec,0,0);
            [theta_Mf_N, fval_Mf_N, Flag_Mf_N] = fmincon(f_Cop_Mf_N,para_start_Mf(1:end-1),[],[],[],[],Lb_Mf(1:end-1),Rb_Mf(1:end-1),[],options);
            para_cell{model_i} = [theta_Mf_N' -fval_Mf_N Flag_Mf_N];
            
            % used to print the results
            f_handle  = f_Cop_Mf_N;
            theta_opt = theta_Mf_N;
            max_LogLik = -fval_Mf_N;
            
        case 8
            % MF t model
            f_Cop_Mf_t = @(params)LogLik_Factor_Copula.LogLik_Copula_MF(N,T,params,u_mat,asset_group_vec,1,0);
            [theta_Mf_t, fval_Mf_t, Flag_Mf_t] = fmincon(f_Cop_Mf_t,para_start_Mf,[],[],[],[],Lb_Mf,Rb_Mf,[],options);
            para_cell{model_i} = [theta_Mf_t' -fval_Mf_t Flag_Mf_t];
            
            % used to print the results
            f_handle  = f_Cop_Mf_t;
            theta_opt = theta_Mf_t;
            max_LogLik = -fval_Mf_t;
            
        case 9
            % MF LT N  model
            
            % 1st step: moment estimator (mimimize eq(12) of Opschoor et al. (2020))
            % NOTE: DATA IS ALREADY BEEN SORTED ACROSS INDUSTRY (from 1 to ..... 10)
            % Compute matrix R_m of eq(12)
            
            x_mat = norminv(u_mat);
            Rt_Block = Admin.Compute_BLOCK_correlation_matrix(x_mat,n_vec);
            Rt_sample_vech = Admin.Sym2Vech(G,Rt_Block');
            
            f_Moment = @(params)LogLik_Factor_Copula.Moment_based_omega_LT_FC(G,params, Rt_sample_vech);
            [f_bar,md_eq_ind,Flag]  = fmincon(f_Moment, para_start_Mf_LT_1,[],[],[],[],Lb_Mf_LT_1,Rb_Mf_LT_1,[],options);
            
            % 2nd step: ML estimation given f_bar
            f_Cop_Mf_LT_N_given_omega = @(params)LogLik_Factor_Copula.LogLik_Copula_LT_factor_given_omega(N,T,params,f_bar,u_mat,asset_group_vec,n_vec,0,0);
            [theta_Mf_LT_N, fval_Mf_LT_N, Flag_Mf_LT_N] = fmincon(f_Cop_Mf_LT_N_given_omega ,para_start_Mf_LT_2(1:end-1),[],[],[],[],Lb_Mf_LT_2(1:end-1),Rb_Mf_LT_2(1:end-1),[],options);
            
            para_cell{model_i}{1} = [theta_Mf_LT_N' -fval_Mf_LT_N Flag_Mf_LT_N];
            para_cell{model_i}{2} = f_bar;
            
            % used to print the results
            f_handle  = f_Cop_Mf_LT_N_given_omega;
            theta_opt = theta_Mf_LT_N;
            max_LogLik = -fval_Mf_LT_N;
            
        case 10
            % MF_LT (t)
            
            % step 1: moment estimator of f_bar (see eq 11/12 of Opschoor et al. (2020))
            % NOTE: DATA IS ALREADY BEEN SORTED ACROSS INDUSTRY (from 1 to ..... 10)
            % (Note also that the moment estimator is the same as in the Gaussian case)
            % Compute matrix R_m of eq(12)
            
            x_mat    = norminv(u_mat);
            Rt_Block = Admin.Compute_BLOCK_correlation_matrix(x_mat,n_vec);
            Rt_sample_vech = Admin.Sym2Vech(G,Rt_Block');
            
            f_Moment = @(params)LogLik_Factor_Copula.Moment_based_omega_LT_FC(G,params, Rt_sample_vech);
            [f_bar,md_eq_ind,Flag]  = fmincon(f_Moment, para_start_Mf_LT_1,[],[],[],[],Lb_Mf_LT_1,Rb_Mf_LT_1,[],options);
            
            % step 2: ML (t copula) given f_bar
            f_Cop_Mf_LT_t_given_omega = @(params)LogLik_Factor_Copula.LogLik_Copula_LT_factor_given_omega(N,T,params,f_bar,u_mat,asset_group_vec,n_vec,1,0);
            [theta_Mf_LT_t, fval_Mf_LT_t, Flag_Mf_LT_t] = fmincon(f_Cop_Mf_LT_t_given_omega ,para_start_Mf_LT_2,[],[],[],[],Lb_Mf_LT_2,Rb_Mf_LT_2,[],options);
            
            para_cell{model_i}{1} = [theta_Mf_LT_t' -fval_Mf_LT_t Flag_Mf_LT_t];
            para_cell{model_i}{2} = f_bar;
            
            % used to print the results
            f_handle  = f_Cop_Mf_LT_t_given_omega;
            theta_opt = theta_Mf_LT_t;
            max_LogLik = -fval_Mf_LT_t;
            
            
    end
    
    nr_para = length(theta_opt);
    
    % compute the standard errors (using the sandwich cov matrix)
    if ind_se==1
        [~,se_S] = Admin.Compute_se_errors_ML(f_handle,theta_opt,T);
        se_para_cell{model_i} = se_S;
    else
        se_S = NaN * ones(nr_para,1);
    end
    
    
    % Print results
    fprintf('Estimated parameters of Model %2i (s.e. in parentheses) \n', model_i);    
     
    for p = 1:nr_para
        fprintf(' %s: %2.4f (%2.4f) \n',para_cell_str{model_i}{p},theta_opt(p),se_S(p));
    end
    
    fprintf('Maximized LogLikelihood: %8.1f \n', max_LogLik)
    disp(' ')
    
end







